use std::pin::Pin;

use futures::Future;
use nostr_rs_client::Relayers;
use nostr_rs_relayer::Relayer;
use nostr_rs_storage::RocksDb;
use nostr_rs_types::{Request, Response};
use tokio::{net::TcpListener, sync::mpsc};

fn on_connection(
    sent_messages: Vec<Request>,
    socket: mpsc::Sender<Request>,
) -> Pin<Box<dyn Future<Output = ()> + Send>> {
    println!("Reconnecting with {:?}", sent_messages);
    Box::pin(async move {
        for m in sent_messages {
            let _ = socket.send(m).await;
        }
    })
}

#[tokio::main]
async fn main() {
    let db = RocksDb::new("./relayer-db").expect("db");
    let (relayer, mut server_receiver) = Relayer::new(db);
    let (mut clients, mut client_receiver) = Relayers::new();
    let _ = clients
        .connect_to("wss://relay.damus.io/", u16::MAX, Some(on_connection))
        .await
        .expect("valid connection");

    let addr = "127.0.0.1:3000";
    let listener = TcpListener::bind(&addr).await.unwrap();

    let relayer_for_recv = relayer.clone();
    let relayer_for_client = relayer.clone();

    tokio::spawn(async move {
        loop {
            let (event, _) = client_receiver.recv().await.unwrap();
            match &event {
                Response::Event(event) => {
                    println!("Sending event: {:?}", event.event);
                    let _ = relayer_for_client.store_and_broadcast(&event.event);
                }
                msg => println!("recv: {:?}", msg),
            }
        }
    });

    tokio::spawn(async move {
        loop {
            let x = relayer_for_recv.recv(&mut server_receiver).await.unwrap();
            println!("Received: {:?}", x);
            if let Some(x) = x {
                clients.send(x).await;
            }
        }
    });

    loop {
        let (stream, _) = listener.accept().await.unwrap();
        let addr = stream.peer_addr().unwrap();

        println!("Client {} connected", addr);

        let _ = relayer.add_connection(stream).await;
    }
}
